from typing import List

from hopeit.apps_visualizer.site.visualization import VisualizationOptions
from hopeit.app.api import event_api
from hopeit.app.events import collector_step, Collector
from hopeit.app.context import EventContext

from hopeit.apps_visualizer.event_stats.collect import get_stats
from hopeit.apps_visualizer.site.visualization import \
    visualization_options, visualization_options_api_args  # noqa: F401
from hopeit.apps_visualizer.apps.events_graph import EventsGraphResult, \
    config_graph, cytoscape_data, runtime_apps  # noqa: F401
from hopeit.apps_visualizer import AppsVisualizerSettings

__steps__ = [
    'visualization_options',
    collector_step(payload=VisualizationOptions).gather(
        'runtime_apps',
        'config_graph',
        'cytoscape_data',
        'live_stats'
    ),
    'build_visualization'
]

__api__ = event_api(
    summary="App Visualizer: Live Stats",
    description="App Visualizer: Live Stats",
    query_args=visualization_options_api_args(),
    responses={
        200: (EventsGraphResult, "Graph Data with applied Live Stats")
    }
)


def _classes(item: dict, new_classes: List[str]) -> str:
    return ' '.join(sorted(
        {*item.get('classes', '').split(' '), *new_classes}
    )).lstrip(' ')


async def live_stats(collector: Collector, context: EventContext) -> Collector:
    settings = context.settings(key='apps_visualizer', datatype=AppsVisualizerSettings)
    apps = await collector['runtime_apps']
    options = await collector['payload']
    host_pids = set(
        (server.host_name, server.pid)
        for _, runtime_info in apps.apps.items()
        for server in runtime_info.servers
        if options.host_filter in server.url
    )
    event_stats = get_stats(
        host_pids=host_pids,
        time_window_secs=settings.live_active_treshold_seconds,
        recent_secs=settings.live_recent_treshold_seconds
    )
    graph = await collector['cytoscape_data']
    if len(event_stats) == 0:
        return graph

    for item_id, item in graph.data.items():
        label = item['data'].get('label', '')
        source = item['data'].get('source', '')
        target = item['data'].get('target', '')
        if label and source and source[0] == '>' and source[-len(label):] != label:
            source += '.' + label
        if target and target[0] == '>':
            target = ''
        keys = filter(bool, [item_id, source, target])

        classes = []
        for key in keys:
            s = event_stats.get(key)
            if s:
                if s.started:
                    classes.append('STARTED')
                if s.pending:
                    classes.append('PENDING')
                if s.recent:
                    classes.append('RECENT')
                if s.failed:
                    classes.append('FAILED')
                if s.ignored:
                    classes.append('IGNORED')
                item['classes'] = _classes(item, classes)
                target = item['data'].get('target', ' ')
                if target[0] == '>':
                    target_item = graph.data[target]
                    target_item['classes'] = _classes(target_item, classes)
                if len(source) > 5 and (
                    source[-5:] == ".POST" or source[-4:] == ".GET" or source[-8:] == "MULTIPART"
                ):
                    source_item = graph.data[source]
                    source_item['classes'] = _classes(source_item, classes)
    return graph


async def build_visualization(collector: Collector, context: EventContext) -> EventsGraphResult:
    return EventsGraphResult(
        runtime_apps=await collector['runtime_apps'],
        graph=await collector['live_stats'],
        options=await collector['payload']
    )



"""
Events graph showing events, stream and dependencies for specified apps
"""
import os
from pathlib import Path

from hopeit.app.context import EventContext, PostprocessHook

from hopeit.server.names import route_name
from hopeit.app.api import event_api
from hopeit.dataobjects import dataclass, dataobject

from hopeit.config_manager import RuntimeApps

from hopeit.apps_visualizer.apps import get_runtime_apps
from hopeit.apps_visualizer.site.visualization import VisualizationOptions, \
    visualization_options, visualization_options_api_args  # noqa: F401

__steps__ = [
    'visualization_options',
    'runtime_apps_config'
]

__api__ = event_api(
    summary="App Visualizer: Site",
    description="[Click here to open Events Graph](/ops/apps-visualizer)",
    query_args=visualization_options_api_args(),
    responses={
        200: (str, "HTML page with Events Graph")
    }
)

_dir_path = Path(os.path.dirname(os.path.realpath(__file__)))


@dataobject
@dataclass
class RuntimeAppsConfig:
    runtime_apps: RuntimeApps
    options: VisualizationOptions


async def runtime_apps_config(options: VisualizationOptions, context: EventContext) -> RuntimeAppsConfig:
    """
    Extract current runtime app_config objects
    """
    return RuntimeAppsConfig(
        runtime_apps=await get_runtime_apps(context, refresh=True, expand_events=options.expanded_view),
        options=options
    )


async def __postprocess__(result: RuntimeAppsConfig, context: EventContext, response: PostprocessHook) -> str:
    """
    Renders html from template, using cytospace data json
    """
    response.set_content_type("text/html")

    app_prefix = result.options.app_prefix

    view_link = f"apps-visualizer?app_prefix={result.options.app_prefix}"
    view_link += f"&host_filter={result.options.host_filter}"
    view_link += f"&expanded_view={str(not result.options.expanded_view).lower()}"
    view_link += f"&live={str(result.options.live).lower()}"

    live_link = f"apps-visualizer?app_prefix={result.options.app_prefix}"
    live_link += f"&host_filter={result.options.host_filter}"
    live_link += f"&expanded_view={str(result.options.expanded_view).lower()}"
    live_link += f"&live={str(not result.options.live).lower()}"

    app_prefix = f"{result.options.app_prefix}*" if result.options.app_prefix else 'All running apps'
    host_filter = f"*{result.options.host_filter}*" if result.options.host_filter else 'All servers'
    view_type = "Effective Events" if result.options.expanded_view else "Configured Events"
    live_type = "Live!" if result.options.live else "Static"

    refresh_endpoint_comps = (
        ['event-stats', 'live'] if result.options.live else ['apps', 'events-graph']
    )
    refresh_endpoint = route_name("api", context.app.name, context.app.version, *refresh_endpoint_comps)
    refresh_endpoint += f"?app_prefix={result.options.app_prefix}"
    refresh_endpoint += f"&host_filter={result.options.host_filter}"
    refresh_endpoint += f"&expanded_view={str(result.options.expanded_view).lower()}"
    refresh_endpoint += f"&live={str(result.options.live).lower()}"

    with open(_dir_path / 'events_graph_template.html') as f:
        template = f.read()
        template = template.replace("{{ app_prefix }}", app_prefix)
        template = template.replace("{{ host_filter }}", host_filter)
        template = template.replace("{{ view_link }}", view_link)
        template = template.replace("{{ live_link }}", live_link)
        template = template.replace("{{ refresh_endpoint }}", refresh_endpoint)
        template = template.replace("{{ view_type }}", view_type)
        template = template.replace("{{ live_type }}", live_type)
        return template



"""
Apps Visualizer: graph elements model
"""
from typing import List, Dict
from dataclasses import dataclass, field
from enum import Enum

from hopeit.dataobjects import dataobject
from hopeit.app.config import AppConnection, AppDescriptor, EventDescriptor, EventType, StreamQueueStrategy
from hopeit.server.names import auto_path


class NodeType(str, Enum):
    REQUEST = "REQUEST"
    EVENT = "EVENT"
    STREAM = "STREAM"
    MULTIPART = "MULTIPART"


@dataobject
@dataclass
class Node:
    id: str
    label: str
    type: NodeType
    inputs: List[str] = field(default_factory=list)
    outputs: List[str] = field(default_factory=list)
    slots: List[str] = field(default_factory=list)


@dataobject
@dataclass
class Edge:
    id: str
    label: str
    source: str
    target: str


@dataobject
@dataclass
class Graph:
    nodes: List[Node]
    edges: List[Edge]


def get_nodes(events: Dict[str, EventDescriptor],
              *, expanded_view: bool = False) -> Dict[str, Node]:
    """
    Create Node metadata from EventDescriptors from app_config,
    expanding effective events using engine functionallity.

    :param expanded_view: bool, if True, stream queues are shown as separate streams,
        otherwise they are shown in a single node with multiple input/outputs.
    """
    nodes = {}
    for event_name, event_info in events.items():
        inputs, outputs = [], []
        if event_info.type in (EventType.GET, EventType.POST, EventType.MULTIPART):
            port_name = f"{event_name}.{event_info.type.value}"
            inputs.append(port_name)
            request_node = Node(
                id=port_name, label=event_info.type.value, type=NodeType.REQUEST, outputs=[port_name]
            )
            nodes[port_name] = request_node

        if event_info.read_stream:
            queues = event_info.read_stream.queues
            for qid, queue in zip(queues, queues) if expanded_view else [("", "|".join(queues))]:
                stream_id = f">{event_info.read_stream.name}.{qid}".strip('.')
                stream_name = f"{event_info.read_stream.name}"
                if qid not in ("", "AUTO"):
                    stream_name += f".{qid}"
                stream_node = nodes.get(stream_id, Node(
                    id=stream_id, label=stream_name, type=NodeType.STREAM
                ))
                stream_node.slots = sorted(set([*stream_node.slots, *queue.split("|")]))
                nodes[stream_id] = stream_node

                for q in [queue] if expanded_view else queues:
                    port_name = f"{event_name}.{stream_name}.{q}"
                    inputs.append(port_name)
                    stream_node.outputs.append(port_name)

        if event_info.write_stream:
            queues = event_info.write_stream.queues
            if event_info.read_stream and event_info.write_stream.queue_strategy == StreamQueueStrategy.PROPAGATE:
                queues = [
                    qx if qy == "AUTO" else qy
                    for qx in event_info.read_stream.queues
                    for qy in queues
                ]
            for qid, queue in zip(queues, queues) if expanded_view else [("", "|".join(queues))]:
                stream_id = f">{event_info.write_stream.name}.{qid}".strip('.')
                stream_name = f"{event_info.write_stream.name}"
                if qid not in ("", "AUTO"):
                    stream_name += f".{qid}"
                stream_node = nodes.get(stream_id, Node(
                    id=stream_id, label=stream_name, type=NodeType.STREAM
                ))
                stream_node.slots = sorted(set([*stream_node.slots, *queue.split("|")]))
                nodes[stream_id] = stream_node

                for q in [queue] if expanded_view else queues:
                    port_name = f"{event_name}.{stream_name}.{q}"
                    stream_node.inputs.append(port_name)
                    outputs.append(port_name)

        nodes[event_name] = Node(
            id=event_name, label=event_name, type=NodeType.EVENT, inputs=inputs, outputs=outputs
        )

    return nodes


def add_app_connections(nodes: Dict[str, Node], *,
                        app_connections: Dict[str, AppConnection],
                        events: Dict[str, EventDescriptor]):
    """
    Add to nodes map, input/output ports for app_connections in order to show call
    dependencies between events.
    """
    for event_name, event_info in events.items():
        app_key = '.'.join(event_name.split('.', maxsplit=2)[0:2])
        source_node = nodes[event_name]
        if source_node:
            for conn in event_info.connections:
                app_connection = app_connections[f"{app_key}.{conn.app_connection}"]
                target_app_key = AppDescriptor(name=app_connection.name, version=app_connection.version).app_key()
                dest_node_name = f"{target_app_key}."
                if app_connection.plugin_name:
                    dest_node_name += auto_path(
                        app_connection.plugin_name,
                        app_connection.plugin_version or app_connection.version
                    ) + '.'
                dest_node_name += f"{conn.event}.{conn.type.value}"
                dest_node = nodes.get(dest_node_name)
                if dest_node:
                    port = f"{event_name}.{app_connection.name}.{conn.event}.{conn.type.value}"
                    source_node.outputs.append(port)
                    dest_node.inputs.append(port)


def get_edges(nodes: Dict[str, Node]):
    """
    Builds Edge list from list of Nodes
    """
    inputs = {
        port: node for node in nodes.values() for port in node.inputs
    }
    outputs = {
        port: node for node in nodes.values() for port in node.outputs
    }

    edges = []
    for k, source in outputs.items():
        target = inputs.get(k)
        if target:
            edges.append(Edge(
                id=k,
                label=k.split(".")[-1],
                source=source.id,
                target=target.id
            ))

    return edges
